from aiohttp import ClientSession

from drukarnia_api.objects.base_object import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2comments

from datetime import datetime
from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:  # always False, used for type hints
    from drukarnia_api.objects.author import Author


class Comment(DrukarniaElement):
    async def reply(self, comment_text: str) -> str:
        """
        Posts a reply to a comment and returns the ID of the new comment.

        Parameters:
            comment_text (str): The text of the comment to be posted as a reply.

        Returns:
            str: The ID of the new comment.
        """

        new_comment_id = await self.request(
            'post',
            f'/api/articles/{await self.article_id}/comments/{await self.comment_id}/replies',
            data={
                "comment": comment_text,
                "replyToComment": await self.comment_id,
                "rootComment": await self.comment_id,   # not sure
                "rootCommentOwner": await self.owner_id,
                "replyToUser": await self.owner_id,
            },
            output='read'
        )

        return new_comment_id.decode('utf-8')

    async def delete(self) -> None:
        """
        Deletes a comment from the article.
        """

        await self.request(
            'delete',
            f'/api/articles/{await self.article_id}/comments/{await self.comment_id}')

    @DrukarniaElement.requires_attributes(['article_id', 'comment_id'])
    async def get_replies(self,
                          create_comments: bool = True) -> Tuple['Comment'] or Tuple[Dict]:
        """
        Deletes a comment from the article.
        """

        replies = await self.request(
            'get',
            f'/api/articles/{await self.article_id}/comments/{await self.comment_id}/replies',
            output='json'
        )

        if create_comments:
            replies = await data2comments(replies, session=self.session)

        return replies

    @DrukarniaElement.requires_attributes(['article_id'])
    async def like_comment(self, unlike: bool = False) -> None:
        """
        Likes or unlikes a comment based on the 'delete' parameter.

        Parameters:
            unlike (bool, optional): If True, the comment will be unliked.
            If False, the comment will be liked. Default is False.
        """

        if unlike:
            await self.request(
                'delete',
                f'/api/articles/{await self.article_id}/comments/{await self.comment_id}/likes')

        else:
            await self.request(
                'post',
                f'/api/articles/{await self.article_id}/comments/{await self.comment_id}/likes')

    @property
    @DrukarniaElement.type_decorator(datetime)
    async def created_at(self) -> datetime:
        """
        Property to access the creation date of the comment.

        Returns:
            datetime: The creation date of the comment.
        """

        return self._access_data('createdAt')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def hidden(self) -> bool:
        """
        Property to check if the comment is hidden by the author.

        Returns:
            bool: True if the comment is hidden, False otherwise.
        """
        return self._access_data('hiddenByAuthor')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def is_blocked(self) -> bool:
        """
        Property to check if the comment is blocked.

        Returns:
            bool: True if the comment is blocked, False otherwise.
        """
        return self._access_data('isBlocked')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def is_liked(self) -> bool:
        """
        Property to check if the comment is liked by you.

        Returns:
            bool: True if the comment is liked, False otherwise.
        """
        return self._access_data('isLiked')

    @property
    @DrukarniaElement.type_decorator(int)
    async def number_of_replies(self) -> int:
        """
        Property to get the number of replies to the comment.

        Returns:
            int: The number of replies to the comment.
        """
        return self._access_data('replyNum')

    @property
    @DrukarniaElement.type_decorator(int)
    async def number_of_likes(self) -> int:
        """
        Property to get the number of likes on the comment.

        Returns:
            int: The number of likes on the comment.
        """
        return self._access_data('likesNum')

    @property
    @DrukarniaElement.type_decorator(str)
    async def article_id(self) -> str or None:
        """
        Property to get the ID of the associated article.

        Returns:
            str or None: The ID of the associated article if available, otherwise None.
        """
        return self._access_data('article')

    @property
    async def owner(self) -> 'Author':
        """
        Property to get the owner (author) of the comment.

        Returns:
            Author: The Author object representing the owner of the comment.
        """
        author = await data2authors([self._access_data('owner', [])], self.session)
        return author[0] if author else None

    @property
    @DrukarniaElement.type_decorator(str)
    async def text(self) -> str:
        """
        Property to get the text of the comment.

        Returns:
            str: The text of the comment.
        """
        return self._access_data('comment')

    @property
    @DrukarniaElement.type_decorator(str)
    async def owner_id(self) -> str or None:
        """
        Property to get the ID of the owner (author) of the comment.

        Returns:
            str or None: The ID of the owner (author) of the comment if available, otherwise None.
        """
        return self._access_data('owner', {}).get('_id', None)

    @property
    @DrukarniaElement.type_decorator(str)
    async def comment_id(self) -> str:
        """
        Retrieves the ID of the comment.

        Returns:
            str: The ID of the comment.
        """
        return self._access_data('_id')

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Comment':
        """
        Creates a Comment instance from records.

        Parameters:
            session (ClientSession): The aiohttp ClientSession object to use for API requests.
            new_data (dict): A dictionary containing the data for the new comment.

        Returns:
            Comment: The newly created Comment instance.
        """
        new_comment = Comment(session=session)
        new_comment._update_data(new_data)

        return new_comment



from datetime import datetime
from typing import Dict, Tuple, List
from aiohttp import ClientSession

from drukarnia_api.objects.base_object import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags

from typing import TYPE_CHECKING

if TYPE_CHECKING:   # always False, used for type hints
    from drukarnia_api.objects.article import Article
    from drukarnia_api.objects.tag import Tag


class Author(DrukarniaElement):
    def __init__(self, username: str = None, author_id: str = None, *args, **kwargs):
        """
        Initialize an Author object.

        Parameters:
            username (str, optional): The username of the author. Defaults to None.
            author_id (str, optional): The ID of the author. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

        self._update_data({'username': username, '_id': author_id})

    async def login(self, email: str, password: str) -> None:
        """
        Log in the author with the provided email and password.

        Parameters:
            email (str): The email of the author.
            password (str): The password of the author.

        Returns:
            None
        """

        await self.cookies.login(email, password, self)
        self._update_data(self.cookies.owner['user'])

    @DrukarniaElement.requires_attributes(['author_id'])
    async def get_followers(self, create_authors: bool = True, offset: int = 0, results_per_page: int = 20,
                            n_collect: int = None, *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get the followers of the author.

        Parameters:
            create_authors (bool, optional): Whether to create Author objects for followers. Defaults to True.
            offset (int, optional): Offset for pagination. Defaults to 0.
            results_per_page (int, optional): Number of results per page for pagination. Defaults to 20.
            n_collect (int, optional): Total number of results to collect. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tuple['Author'] or Tuple[Dict]: A tuple of Author objects or dictionaries representing the followers.
        """

        # Make a request to get the followers of the author
        followers = await self.multi_page_request(f'/api/relationships/{await self.author_id}/followers',
                                                  offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            followers = await data2authors(followers, self.session)

        return followers

    @DrukarniaElement.requires_attributes(['author_id'])
    async def get_followings(self, create_authors: bool = True, offset: int = 0, results_per_page: int = 20,
                             n_collect: int = None, *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get the followings of the author.

        Parameters:
            create_authors (bool, optional): Whether to create Author objects for followings. Defaults to True.
            offset (int, optional): Offset for pagination. Defaults to 0.
            results_per_page (int, optional): Number of results per page for pagination. Defaults to 20.
            n_collect (int, optional): Total number of results to collect. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tuple['Author'] or Tuple[Dict]: A tuple of Author objects or dictionaries representing the followings.
        """

        # Make a request to get the followings of the author
        followings = await self.multi_page_request(f'/api/relationships/{await self.author_id}/following',
                                                   offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            followings = await data2authors(followings, self.session)

        return followings

    async def get_notifications(self, offset: int = 0, results_per_page: int = 20,
                                n_collect: int = None, *args, **kwargs) -> List[Dict]:
        """
        Get the notifications of the author.

        Parameters:
            offset (int, optional): Offset for pagination. Defaults to 0.
            results_per_page (int, optional): Number of results per page for pagination. Defaults to 20.
            n_collect (int, optional): Total number of results to collect. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            List[Dict]: A list of dictionaries representing the notifications.
        """
        return await self.multi_page_request('/api/notifications',
                                             offset, results_per_page, n_collect,
                                             *args, **kwargs)

    async def get_reads_history(self, create_articles: bool = True, offset: int = 0,
                                results_per_page: int = 20, n_collect: int = None,
                                *args, **kwargs) -> List[Dict] or List['Article']:
        """
        Get the reading history of the author.

        Parameters:
            create_articles (bool, optional): Whether to create Article objects for read history. Defaults to True.
            offset (int, optional): Offset for pagination. Defaults to 0.
            results_per_page (int, optional): Number of results per page for pagination. Defaults to 20.
            n_collect (int, optional): Total number of results to collect. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            List[Dict] or List['Article']: A list of dictionaries or Article objects representing the reading history.
        """
        articles = await self.multi_page_request('/api/stats/reads/history',
                                                 offset, results_per_page, n_collect,
                                                 *args, **kwargs)

        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    async def get_sections(self, preview: bool = True, **kwargs) -> List[dict]:
        """
        Get the sections of the author's articles.

        Parameters:
            preview (bool, optional): Whether to get sections with preview. Defaults to True.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            List[dict]: A list of dictionaries representing the sections.
        """
        return await self.request('get', f'/api/articles/bookmarks/lists?preview={str(preview).lower()}',
                                  output='json', **kwargs)

    async def create_section(self, name: str, **kwargs) -> List[Dict]:
        """
        Create a new section for the author's articles.

        Parameters:
            name (str): The name of the new section.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            List[Dict]: A list of dictionaries representing the new section.
        """
        section_id = await self.request('get',
                                        '/api/articles/bookmarks/lists',
                                        data={"name": name},
                                        output='read', **kwargs)

        return section_id.decode('utf-8')

    async def delete_section(self, section_id: str, **kwargs) -> None:
        """
        Delete a section for the author's articles.

        Parameters:
            section_id (str): The ID of the section to be deleted.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        await self.request('delete', f'/api/articles/bookmarks/lists/{section_id}', **kwargs)

    @DrukarniaElement.requires_attributes(['author_id'])
    async def subscribe(self, unsubscribe: bool = False) -> None:
        """
        Subscribe or unsubscribe to/from an author.

        Parameters:
            unsubscribe (bool, optional): Whether to unsubscribe. Defaults to False.

        Returns:
            None
        """

        if unsubscribe:
            await self.request('delete', f'/api/relationships/subscribe/{await self.author_id}')
            return None

        await self.request('post', f'/api/relationships/subscribe/{await self.author_id}')

    @DrukarniaElement.requires_attributes(['author_id'])
    async def block(self, unblock: bool = False) -> None:
        """
        Block or unblock an author.

        Parameters:
            unblock (bool, optional): Whether to unblock. Defaults to False.

        Returns:
            None
        """

        if unblock:
            await self.request('patch', f'/api/relationships/block/{await self.author_id}')
            return None

        await self.request('put', f'/api/relationships/block/{await self.author_id}')

    async def get_blocked(self, create_authors: bool = False) -> List[Dict] or Tuple['Author']:
        """
        Get the authors blocked by the current author.

        Parameters:
            create_authors (bool, optional): Whether to create Author objects for blocked authors. Defaults to False.

        Returns:
            List[Dict] or Tuple['Author']: A list of dictionaries or Author objects representing the blocked authors.
        """

        authors = await self.request('get', '/api/relationships/blocked', output='json')

        if create_authors:
            authors = await data2authors(authors, self.session)

        return authors

    async def change_password(self, old_password: str, new_password: str, **kwargs) -> None:
        """
        Change the author's password.

        Parameters:
            old_password (str): The current password of the author.
            new_password (str): The new password for the author.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        await self.request('patch', f'/api/users/login/password',
                           data={"oldPassword": old_password, "newPassword": new_password},
                           output='read', **kwargs)

    async def change_user_info(self, name: str = None, description: str = None, username: str = None,
                               description_short: str = None, socials: dict = None, donate_url: str = None) -> str:
        """
        Change the author's user information.

        Parameters:
            name (str, optional): The new name of the author. Defaults to None.
            description (str, optional): The new description of the author. Defaults to None.
            username (str, optional): The new username of the author. Defaults to None.
            description_short (str, optional): The new short description of the author. Defaults to None.
            socials (dict, optional): The new socials information of the author. Defaults to None.
            donate_url (str, optional): The new donate URL of the author. Defaults to None.

        Returns:
            str: The response received from the server.
        """
        info2patch = {"name": name, "description": description, "username": username,
                      "descriptionShort": description_short,
                      "socials": socials, "donateUrl": donate_url}

        info2patch = {key: value for key, value in info2patch.items() if value is not None}

        response = await self.request('patch', '/api/users', data=info2patch, output='read')
        return response.decode('utf-8')

    async def change_email(self, current_password: str, new_email: str, **kwargs) -> None:
        """
        Change the author's email.

        Parameters:
            current_password (str): The current password of the author.
            new_email (str): The new email for the author.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        await self.request('patch', f'/api/users/login/email',
                           data={"currentPassword": current_password, "newEmail": new_email},
                           **kwargs)

    @DrukarniaElement.requires_attributes(['username'], solution='provide username while initializing Author.')
    async def collect_data(self, return_: bool = False) -> Dict or None:
        """
        Collect the author's data and update the object's attributes.

        Parameters:
            return_ (bool, optional): Whether to return the collected data. Defaults to False.

        Returns:
            Dict or None: A dictionary representing the author's data if return_ is True, otherwise None.
        """

        data = await self.request('get', '/api/users/profile/{username}'.format(username=await self.username),
                                  output='json')

        self._update_data(data)

        if return_:
            return data

    @property
    @DrukarniaElement.type_decorator(str)
    async def username(self) -> str:
        """
        Get the username of the author.

        Returns:
            str: The username of the author.
        """
        return self._access_data('username')

    @property
    @DrukarniaElement.type_decorator(str)
    async def avatar(self) -> str:
        """
        Get the avatar of the author.

        Returns:
            str: The avatar URL of the author.
        """
        return self._access_data('avatar')

    @property
    @DrukarniaElement.type_decorator(str)
    async def donate_url(self) -> str:
        """
        Get the donate URL of the author.

        Returns:
            str: The donate URL of the author.
        """
        return self._access_data('donateUrl')

    @property
    @DrukarniaElement.type_decorator(str)
    async def socials(self) -> Dict:
        """
        Get the socials information of the author.

        Returns:
            Dict: A dictionary representing the socials information of the author.
        """
        return self._access_data('socials')

    @property
    @DrukarniaElement.type_decorator(str)
    async def author_id(self) -> str:
        """
        Get the ID of the author.

        Returns:
            str: The ID of the author.
        """
        return self._access_data('_id')

    @property
    @DrukarniaElement.type_decorator(str)
    async def name(self) -> str:
        """
        Get the name of the author.

        Returns:
            str: The name of the author.
        """
        return self._access_data('name')

    @property
    @DrukarniaElement.type_decorator(str)
    async def description(self) -> str:
        """
        Get the description of the author.

        Returns:
            str: The description of the author.
        """
        return self._access_data('description')

    @property
    @DrukarniaElement.type_decorator(str)
    async def description_short(self) -> str:
        """
        Get the short description of the author.

        Returns:
            str: The short description of the author.
        """
        return self._access_data('descriptionShort')

    @property
    @DrukarniaElement.type_decorator(datetime)
    async def created_at(self) -> datetime:
        """
        Get the creation date of the author.

        Returns:
            datetime: The creation date of the author.
        """
        return self._access_data('createdAt')

    @property
    @DrukarniaElement.type_decorator(int)
    async def following_num(self) -> int:
        """
        Get the number of authors the current author is following.

        Returns:
            int: The number of authors the current author is following.
        """
        return self._access_data('followingNum')

    @property
    @DrukarniaElement.type_decorator(int)
    async def followers_num(self) -> int:
        """
        Get the number of followers of the author.

        Returns:
            int: The number of followers of the author.
        """
        return self._access_data('followersNum')

    @property
    @DrukarniaElement.type_decorator(int)
    async def read_num(self) -> int:
        """
        Get the number of reads by the author.

        Returns:
            int: The number of reads by the author.
        """
        return self._access_data('readNum')

    @property
    async def articles(self) -> Tuple['Article']:
        """
        Get the articles written by the author.

        Returns:
            Tuple['Article']: A tuple of Article objects representing the articles written by the author.
        """
        return await data2articles(self._access_data('articles', []), self.session)

    @property
    async def author_tags(self) -> Tuple['Tag']:
        """
        Get the tags associated with the author.

        Returns:
            Tuple['Tag']: A tuple of Tag objects representing the tags associated with the author.
        """
        return await data2tags(self._access_data('authorTags', []), self.session)

    @property
    @DrukarniaElement.type_decorator(dict)
    async def relationships(self) -> Dict:
        """
        Get the relationships of the author.

        Returns:
            Dict: A dictionary representing the relationships of the author.
        """
        return self._access_data('relationships')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def is_subscribed(self) -> bool:
        """
        Get the subscription state.

        Returns:
            Bool: True if you follow this author, otherwise False
        """
        return self._access_data('relationships', {}).get('isSubscribed', None)

    @property
    @DrukarniaElement.type_decorator(bool)
    async def is_blocked(self) -> bool:
        """
        Get the blocked state.

        Returns:
            Bool: True if you blocked this author, otherwise False
        """
        return self._access_data('relationships', {}).get('isBlocked', None)

    @staticmethod
    async def from_records(session: ClientSession, new_data: Dict) -> 'Author':
        """
        Create an Author instance from records.

        Parameters:
            session (ClientSession): The aiohttp client session.
            new_data (dict): The data to create the Author instance.

        Returns:
            Author: An Author instance.
        """

        new_author = Author(session=session)
        new_author._update_data(new_data)

        return new_author

    def __hash__(self) -> int:
        """
        Calculate the hash value for the Author object.

        Returns:
            int: The hash value.
        """
        return hash(self.author_id or self.username)



from drukarnia_api.objects.base_object import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags

from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:
    from drukarnia_api.objects.article import Article
    from drukarnia_api.objects.tag import Tag
    from drukarnia_api.objects.author import Author


class Search(DrukarniaElement):
    async def find_author(self,
                          query: str,
                          create_authors: bool = True,
                          with_relations: bool = False,
                          offset: int = 0,
                          results_per_page: int = 20,
                          n_collect: int = None,
                          *args, **kwargs) -> Tuple['Author'] or Tuple[Dict]:
        """
        Search for authors.

        Parameters:
        - query (str): The search query for authors.
        - create_authors (bool, optional): Whether to create author objects from the search results. Defaults to True.
        - with_relations (bool, optional): Whether to include author relationships in the search results.
         Defaults to False.
        - offset (int, optional): The starting index of the search results. Defaults to 0.
        - results_per_page (int, optional): The maximum number of authors to be returned per page. Defaults to 20.
        - n_collect (int, optional): The total number of authors to be collected. Defaults to None.

        Returns:
        - Tuple['Author'] or Tuple[Dict]: A tuple containing a list of
        Author objects or a dictionary representing the search results.
        """

        with_relations = str(with_relations).lower()

        # Make a request to get authors
        authors = await self.multi_page_request(f'/api/users/info?name={query}&withRelationships={with_relations}',
                                                offset, results_per_page, n_collect, *args, **kwargs)

        if create_authors:
            authors = await data2authors(authors, self.session)

        return authors

    async def find_articles(self,
                            query: str,
                            create_articles: bool = True,
                            offset: int = 0,
                            results_per_page: int = 20,
                            n_collect: int = None,
                            *args, **kwargs) -> Tuple['Article'] or Tuple[Dict]:
        """
        Search for articles.

        Parameters:
        - query (str): The search query for articles.
        - create_articles (bool, optional): Whether to create article objects from the search results. Defaults to True.
        - offset (int, optional): The starting index of the search results. Defaults to 0.
        - results_per_page (int, optional): The maximum number of articles to be returned per page. Defaults to 20.
        - n_collect (int, optional): The total number of articles to be collected. Defaults to None.

        Returns:
        - Tuple['Article'] or Tuple[Dict]: A tuple containing a list of
        Article objects or a dictionary representing the search results.
        """

        # Make a request to get articles
        articles = await self.multi_page_request(f'/api/articles/search?name={query}',
                                                 offset, results_per_page, n_collect, *args, **kwargs)

        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    async def find_tags(self,
                        query: str,
                        create_tags: bool = True,
                        offset: int = 0,
                        results_per_page: int = 20,
                        n_collect: int = None, **kwargs) -> Tuple['Tag'] or Tuple[Dict]:
        """
        Search for tags.

        Parameters:
        - query (str): The search query for tags.
        - create_tags (bool, optional): Whether to create tag objects from the search results. Defaults to True.
        - offset (int, optional): The starting index of the search results. Defaults to 0.
        - results_per_page (int, optional): The maximum number of tags to be returned per page. Defaults to 20.
        - n_collect (int, optional): The total number of tags to be collected. Defaults to None.
        - **kwargs: Additional keyword arguments to pass to the search API.

        Returns:
        - Tuple['Tag'] or Tuple[Dict]: A tuple containing a list of Tag objects or
         a dictionary representing the search results.
        """

        # Make a request to get articles
        tags = await self.multi_page_request(
            f'/api/articles/search/tags?text={query}',
            key='tags',
            offset=offset,
            results_per_page=results_per_page,
            n_collect=n_collect, **kwargs
        )

        if create_tags:
            tags = await data2tags(tags, self.session)

        return tags



from typing import Dict
from fake_useragent import UserAgent


class Headers(dict):
    def __init__(self,  **kwargs):
        """
        Initialize the Headers object.

        Parameters:
            dynamic_user_generation (bool):
               If True, enables dynamic generation of the User-Agent header using fake_useragent.
               If False, uses a static User-Agent header generated once during initialization.
               Defaults to False.
            **kwargs: Additional headers as key-value pairs.

        Returns:
            None
        """
        super().__init__(**{
            'Content-Type': 'application/json',
            'User-Agent': UserAgent().random,
        })

        self.update(kwargs)



from datetime import datetime
from typing import Any, Callable, TypeVar, List, Awaitable
from attrdict import AttrDict
from drukarnia_api.network.connection import Connection


T = TypeVar('T')


def _synthesizer(offset: int, results_per_page: int, direct_url: str, kwargs: dict):
    """
    Generates requests with pagination.

    Parameters:
        offset (int): The starting offset for pagination.
        results_per_page (int): The number of results per page.
        direct_url (str): The base URL for the requests.
        kwargs (dict): Additional keyword arguments for the requests.

    Yields:
        dict: A dictionary containing request parameters.
    """
    start_page = offset // results_per_page + 1

    while True:
        yield {'url': direct_url,
               'params': {'page': start_page},
               'output': 'json',
               'method': 'get'} | kwargs

        start_page += 1


def _to_datetime(date: str) -> datetime:
    """
    Convert a string representation of a date to a datetime object.

    Parameters:
        date (str): The date string in ISO format.

    Returns:
        datetime: The converted datetime object.
    """
    if date:
        date = datetime.fromisoformat(date[:-1])

    return date


class DrukarniaElement(Connection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data = AttrDict({})

    def _update_data(self, new_data: dict):
        """
        Update the data properties of the DrukarniaElement.

        Parameters:
            new_data (dict): The new data to update the properties with.
        """
        self.data.update(new_data)

    def _access_data(self, key: str, default: Any = None) -> Any:
        """
        Access data from the properties of the DrukarniaElement.

        Parameters:
            key (str): The key to access the data.
            default (Any, optional): The default value to return if the key is not found. Defaults to None.

        Returns:
            Any: The value associated with the given key or the default value if the key is not found.
        """
        return self.data.get(key, default)

    async def multi_page_request(self, direct_url: str, offset: int = 0, results_per_page: int = 20,
                                 n_collect: int = None, key: str = None, **kwargs) -> List:
        """
        Perform a multi-page request with pagination.

        Parameters:
            direct_url (str): The base URL for the requests.
            offset (int, optional): The starting offset for pagination. Defaults to 0.
            results_per_page (int, optional): The number of results per page. Defaults to 20.
            n_collect (int, optional): The total number of results to collect. Defaults to None.
            key (str, optional): The key to extract records from the returned data. Defaults to None.
            **kwargs: Additional keyword arguments for the requests.

        Returns:
            List: A list of records extracted from the paginated data.
        """
        assert offset >= 0, 'Offset must be greater than or equal to zero.'
        assert (n_collect is None) or (n_collect >= 1), 'n_collect must be greater than or equal to one.'

        n_results = (n_collect // results_per_page + int(n_collect % results_per_page != 0)) if n_collect else None

        data = await self.run_until_no_stop(
            request_synthesizer=_synthesizer(offset, results_per_page, direct_url, kwargs),
            not_stop_until=lambda result: result != [],
            n_results=n_results)

        if key is None:
            records = [record for page in data for record in page]
        else:
            records = [record for page in data for record in page.get(key, [])]

        adjusted_start = offset % results_per_page

        if n_collect:
            return records[adjusted_start:adjusted_start + n_collect]

        return records[adjusted_start:]

    @staticmethod
    def type_decorator(return_type: type) -> Callable[[Callable[..., T]], Callable[..., T]]:
        """
        A decorator to enforce the return type of a method.

        Parameters:
            return_type (type): The expected return type.

        Returns:
            Callable: The decorated method with enforced return type.
        """
        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            async def wrapper(*args, **kwargs) -> T:
                result = await func(*args, **kwargs)
                if result is None: return

                if return_type is datetime:
                    return _to_datetime(str(result))

                return return_type(result)

            return wrapper

        return decorator

    @staticmethod
    def requires_attributes(attrs: List[str], solution: str = 'await collect_data() before.') -> Callable:
        """
        A decorator to ensure that certain attributes are present before executing a method.

        Parameters:
            attrs (List[str]): A list of attribute names that must be present.
            solution (str, optional): A suggestion for a possible solution. Defaults to 'await collect_date() before.'.

        Returns:
            Callable: The decorated method with attribute requirement checks.
        """
        def decorator(func: Callable[..., Awaitable]):
            async def wrapper(self_instance, *args, **kwargs):
                if any([(await getattr(self_instance, attr)) is None for attr in attrs]):
                    raise ValueError(f'This function requires attributes {attrs}, '
                                     f'which are missing. Possible solutions: {solution}')

                return await func(self_instance, *args, **kwargs)

            return wrapper

        return decorator



import asyncio

from aiohttp import ClientSession
from typing import Any, Callable, Dict, Generator, Tuple, List, Iterable

from drukarnia_api.network.utils import to_json, _from_response
from drukarnia_api.network.cookie import DrukarniaCookies
from drukarnia_api.network.headers import Headers


class Connection:
    base_url = 'https://drukarnia.com.ua'

    def __init__(self, headers: Headers = None, cookies: DrukarniaCookies = None, session: ClientSession = None):
        """
        Initialize a Connection object.

        Parameters:
            headers (Headers, optional): Headers to be used for the requests. Defaults to None.
            cookies (DrukarniaCookies, optional): Cookie jar to be used for the requests. Defaults to None.
            session (ClientSession, optional): Custom aiohttp ClientSession to be used for the requests.
            Defaults to None.
        """
        self.cookies = DrukarniaCookies.sync_from_aiohttp_session(session) if session else cookies
        self.headers = headers if headers else Headers()

        self.session = session
        self.custom_session = session is not None

    async def __aenter__(self) -> None:
        """
        Context manager entry method for the Connection.

        Returns:
            Connection: The Connection instance.
        """

        if self.session and (not self.session.closed):
            return

        self.cookies = (cookies := self.cookies if self.cookies else DrukarniaCookies())

        self.session = ClientSession(
            base_url=self.base_url,
            cookies=cookies,
            headers=self.headers
        )

        self.custom_session = False

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Context manager exit method for the Connection.

        Parameters:
            exc_type: Exception type.
            exc_val: Exception value.
            exc_tb: Exception traceback.
        """

        if self.session and (self.custom_session is False):
            await self.session.close()
            self.session = None

    async def request(self, method: str, url: str, output: str or list = None, **kwargs) -> Any:
        """
        Perform an HTTP request using the aiohttp ClientSession.

        Parameters:
            method (str): HTTP method for the request (e.g., 'GET', 'POST', 'PUT', 'PATCH', 'DELETE').
            url (str): The URL to which the request will be sent.
            output (str or list, optional): Expected output format, either 'json' or a list of fields. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to aiohttp session.request.

        Returns:
            Any: The result of the HTTP request, parsed based on the provided output format.
        """

        if (self.session is None) or self.session.closed:
            raise ValueError(f'Trying to make request from an unopened/closed session.\n'
                             f'Solutions |: initialize object with a session.\n'
                             f'Solution || (recommended): use context manager as follows ```async with '
                             f'{self.__class__.__name__}(*args, **kwargs):```')

        if method.lower() in ['post', 'put', 'patch', 'delete']:
            kwargs['data'] = await to_json(kwargs.get('data', {}))

        async with self.session.request(method.upper(), url, **kwargs) as response:
            return await _from_response(response, output)

    async def request_pool(self, heuristics: Iterable[Dict[str, Any]]) -> Tuple:
        """
        Perform a pool of HTTP requests using the given heuristics.

        Parameters:
            heuristics (Iterable[Dict[str, Any]]): An iterable of request heuristics (method, url, output, **kwargs).

        Returns:
            Tuple: A tuple containing the results of all the HTTP requests made in the pool.
        """

        # Create tasks
        tasks = [self.request(**kwargs) for kwargs in heuristics]

        # Get results
        return await asyncio.gather(*tasks)

    async def run_until_no_stop(self, request_synthesizer: Generator[Dict or List[Dict, str], None, None],
                                not_stop_until: Callable[[Any], bool], n_results: int = None,
                                batch_size: int = 5) -> List[Any]:
        """
        Run HTTP requests until a stopping condition is met.

        Parameters:
            request_synthesizer (Generator[Dict or List[Dict, str], None, None]):
                A generator that synthesizes request heuristics (method, url, output, **kwargs).
            not_stop_until (Callable[[Any], bool]):
                A callable that takes a response and returns True if the requests should continue, False otherwise.
            n_results (int, optional): The maximum number of results to fetch. Defaults to None (no limit).
            batch_size (int, optional): The number of requests to send in each batch. Defaults to 5.

        Returns:
            List[Any]: A list containing the results of the HTTP requests that meet the stopping condition.
        """

        all_results = []
        step = 0

        while True:
            heuristics = [next(request_synthesizer) for _ in range(step, step + batch_size)]

            if n_results is not None:
                heuristics = heuristics[:n_results]
                n_results -= batch_size

            _results = await self.request_pool(heuristics=heuristics)
            responses = [_result for _result in _results if not_stop_until(_result)]

            all_results.extend(responses)
            step += batch_size

            if len(responses) != batch_size:
                break

        return all_results



from typing import Tuple
from aiohttp import ClientSession
import asyncio

from typing import TYPE_CHECKING
if TYPE_CHECKING:  # always False, used for type hints
    from drukarnia_api.objects.author import Author
    from drukarnia_api.objects.article import Article
    from drukarnia_api.objects.tag import Tag
    from drukarnia_api.objects.comment import Comment


async def data2tags(tags_data: list or None, session: ClientSession) -> Tuple['Tag'] or Tuple:
    """
    Converts a list of tag data into Tag objects asynchronously.

    Args:
        tags_data (list or None): List of tag data. If None, an empty tuple will be returned.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple[Tag] or Tuple: Tuple of Tag objects or an empty tuple if tags_data is None.
    """
    if not tags_data:
        return ()

    from drukarnia_api.objects.tag import Tag

    tasks = [Tag.from_records(session, tag) for tag in tags_data]

    return await asyncio.gather(*tasks)


async def data2authors(authors_data: list or None, session: ClientSession) -> Tuple['Author'] or Tuple:
    """
    Converts a list of author data into Author objects asynchronously.

    Args:
        authors_data (list or None): List of author data. If None, an empty tuple will be returned.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple[Author] or Tuple: Tuple of Author objects or an empty tuple if authors_data is None.
    """
    if not authors_data:
        return ()

    from drukarnia_api.objects.author import Author

    tasks = [Author.from_records(session, author) for author in authors_data]

    return await asyncio.gather(*tasks)


async def data2articles(articles_data: list or None, session: ClientSession) -> Tuple['Article'] or Tuple:
    """
    Converts a list of article data into Article objects asynchronously.

    Args:
        articles_data (list or None): List of article data. If None, an empty tuple will be returned.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple[Article] or Tuple: Tuple of Article objects or an empty tuple if articles_data is None.
    """
    if not articles_data:
        return ()

    from drukarnia_api.objects.article import Article

    tasks = [Article.from_records(session, article) for article in articles_data]

    return await asyncio.gather(*tasks)


async def data2comments(comment_data: list or None, session: ClientSession) -> Tuple['Comment'] or Tuple:
    """
    Converts a list of comment data into Comment objects asynchronously.

    Args:
        comment_data (list or None): List of comment data. If None, an empty tuple will be returned.
        session (ClientSession): aiohttp ClientSession object.

    Returns:
        Tuple[Comment] or Tuple: Tuple of Comment objects or an empty tuple if comment_data is None.
    """
    if not comment_data:
        return ()

    from drukarnia_api.objects.comment import Comment

    tasks = [Comment.from_records(session, comment) for comment in comment_data]

    return await asyncio.gather(*tasks)



from aiohttp.cookiejar import SimpleCookie
from aiohttp import ClientSession
import re
from typing import TYPE_CHECKING, Tuple, Dict, Any

if TYPE_CHECKING:
    from drukarnia_api.network.connection import Connection


async def _login(email: str, password: str, connection: 'Connection') -> Tuple[str, Dict[str, Any]]:
    """
    Logs in the user by making a POST request to the login endpoint.

    Parameters:
        email (str): The email address of the user.
        password (str): The password of the user.
        connection ('Connection'): An instance of the 'Connection' class used to make HTTP requests.

    Returns:
        Tuple[str, Dict[str, Any]]: A tuple containing two values:
            - str: A string representing the cookies obtained after successful login.
            - Dict[str, Any]: A dictionary containing the response JSON data from the login endpoint.
    """
    headers, info = await connection.request(
        'post',
        '/api/users/login',
        data={"password": password, "email": email},
        output=['headers', 'json']
    )

    headers = str(headers)
    token = re.search(r'refreshToken=(.*?);', headers).group(1)
    device_id = re.search(r'deviceId=(.*?);', headers).group(1)

    cookies = f'deviceId={device_id}; token={token};'

    return cookies, info


class DrukarniaCookies(SimpleCookie):
    def __init__(self, *args, **kwargs):
        """
        Initializes the DrukarniaCookies class, a subclass of CookieJar.

        The unsafe parameter is set to True to allow third-party cookies.
        """
        super().__init__(*args, **kwargs)

        self.owner = None
        self.__authenticated = False

    async def login(self, email: str, password: str, connection: 'Connection'):
        """
        Logs in the user and stores the cookies in the instance.

        Parameters:
            email (str): The email address of the user.
            password (str): The password of the user.
            connection ('Connection'): An instance of the 'Connection' class used to make HTTP requests.
        """

        cookie_str, self.owner = await _login(email, password, connection)
        self.load(cookie_str)

        self.__authenticated = True

    @property
    def authenticated(self) -> bool:
        return self.__authenticated

    @staticmethod
    def sync_from_aiohttp_session(session: ClientSession) -> 'DrukarniaCookies':
        return DrukarniaCookies(session.cookie_jar.filter_cookies('/'))



from datetime import datetime
from warnings import warn
from aiohttp import ClientSession

from drukarnia_api.objects.base_object import DrukarniaElement
from drukarnia_api.shortcuts import data2authors, data2articles, data2tags, data2comments

from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:  # always False, used for type hints
    from drukarnia_api.objects.author import Author
    from drukarnia_api.objects.tag import Tag
    from drukarnia_api.objects.comment import Comment


class Article(DrukarniaElement):
    def __init__(self, slug: str = None, article_id: str = None, *args, **kwargs):
        """
        Initializes an Article object with the given slug and article ID.

        Parameters:
        - slug (str): The slug of the article.
        - article_id (str): The ID of the article.
        """

        super().__init__(*args, **kwargs)

        self._update_data({'slug': slug, '_id': article_id})

    @DrukarniaElement.requires_attributes(['article_id'])
    async def post_comment(self, comment_text: str) -> str:
        """
        Posts a comment on the article and returns the ID of the posted comment.

        Parameters:
        - comment_text (str): The text of the comment to be posted.

        Returns:
        - str: The ID of the posted comment.
        """

        posted_comment_id = await self.request(
            'post',
            '/api/articles/{_id}/comments'.format(_id=await self.article_id),
            data={'comment': comment_text}, output='read')

        return str(posted_comment_id)

    @DrukarniaElement.requires_attributes(['article_id'])
    async def like_article(self, n_likes: int) -> None:
        """
        Likes the article with the specified number of likes.

        Parameters:
        - n_likes (int): The number of likes to be added to the article.

        Raises:
        - ValueError: If the number of likes is not between 0 and 10 (inclusive).
        """

        assert 0 < n_likes <= 10, 'Number of likes must be greater or equal to zero and lower or equal to ten'

        await self.request(
            'post',
            f'/api/articles/{await self.article_id}/like',
            data={'likes': int(n_likes)})

    @DrukarniaElement.requires_attributes(['article_id'])
    async def bookmark(self, section_id: str, unbookmark: bool = False) -> None:
        """
        Adds or removes the article from bookmarks based on the 'unbookmark' parameter.

        Parameters:
        - section_id (str): The ID of the section to bookmark the article.
        - unbookmark (bool): If True, removes the article from bookmarks. If False, adds it to bookmarks.

        Raises:
        - ValueError: If section_id is not provided for bookmarking.
        """

        if unbookmark:
            await self.request('delete', f'/api/articles/{await self.article_id}/bookmarks')

        elif not section_id:
            raise ValueError('section_id must be passed for bookmarking')

        else:
            await self.request('post', '/api/articles/bookmarks',
                               data={"article": await self.article_id, "list": section_id})

    @DrukarniaElement.requires_attributes(['slug'])
    async def collect_data(self, return_: bool = False) -> Dict or None:
        """
        Collects the article's data and updates the object's attributes.

        Parameters:
        - return_ (bool): If True, returns the collected data.

        Returns:
        - Dict or None: The collected data if 'return_' is True, otherwise None.
        """

        data = await self.request('get', f'/api/articles/{await self.slug}', output='json')

        self._update_data(data)

        if return_:
            return data

    @property
    async def owner(self) -> 'Author' or None:
        """
        Retrieves the owner of the article.

        Returns:
        - 'Author' or None: The owner of the article if available, otherwise None.
        """

        owner = self._access_data('owner', None)
        if owner is None:
            return None

        return await data2authors([owner], self.session)

    @property
    async def comments(self) -> Tuple['Comment']:
        """
        Retrieves the comments of the article.

        Returns:
        - Tuple['Comment']: A tuple containing the comments of the article.
        """

        return await data2comments(self._access_data('comments', []), self.session)

    @property
    async def recommended_articles(self) -> Tuple['Article']:
        """
        Retrieves the recommended articles related to the article.

        Returns:
        - Tuple['Article']: A tuple containing the recommended articles.
        """

        return await data2articles(self._access_data('recommendedArticles', []), self.session)

    @property
    async def author_articles(self) -> Tuple['Article']:
        """
        Retrieves the articles written by the author of the article.

        Returns:
        - Tuple['Article']: A tuple containing the articles written by the author.
        """

        return await data2articles(self._access_data('authorArticles', []), self.session)

    @property
    @DrukarniaElement.type_decorator(dict)
    async def relationships(self) -> Dict:
        """
        Retrieves the relationships of the article.

        Returns:
        - Dict: The relationships of the article.
        """

        return self._access_data('relationships')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def is_bookmarked(self) -> bool:
        """
        Checks if the article is bookmarked.

        Returns:
        - bool: True if the article is bookmarked, False otherwise.
        """

        return self._access_data('isBookmarked')

    @property
    @DrukarniaElement.type_decorator(int)
    async def my_likes(self) -> int:
        """
        Checks if the article is liked.

        Returns:
        - bool: True if the article is liked, False otherwise.
        """

        return self._access_data('isLiked')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def sensitive(self) -> bool:
        """
        Checks if the article is sensitive.

        Returns:
        - bool: True if the article is sensitive, False otherwise.
        """

        return self._access_data('sensitive')

    @property
    @DrukarniaElement.type_decorator(dict)
    async def content(self) -> Dict:
        """
        Retrieves the content of the article.

        Returns:
        - Dict: The content of the article.
        """

        return self._access_data('content')

    @property
    @DrukarniaElement.type_decorator(str)
    async def thumb_picture(self) -> str:
        """
        Retrieves the thumbnail picture of the article.

        Returns:
        - str: The URL of the thumbnail picture.
        """

        return self._access_data('thumbPicture')

    @property
    @DrukarniaElement.type_decorator(str)
    async def picture(self) -> str:
        """
        Retrieves the picture of the article.

        Returns:
        - str: The URL of the picture.
        """

        return self._access_data('picture')

    @property
    @DrukarniaElement.type_decorator(str)
    async def ads(self) -> str:
        """
        Retrieves the ads of the article.

        Returns:
        - str: The ads of the article.
        """

        return self._access_data('ads')

    @property
    @DrukarniaElement.type_decorator(str)
    async def index(self) -> str:
        """
        Retrieves the index of the article.

        Returns:
        - str: The index of the article.
        """

        return self._access_data('index')

    @property
    @DrukarniaElement.type_decorator(datetime)
    async def created_at(self) -> datetime:
        """
        Retrieves the creation date of the article.

        Returns:
        - datetime: The creation date of the article.
        """

        return self._access_data('createdAt')

    @property
    @DrukarniaElement.type_decorator(float)
    async def read_time(self) -> float:
        """
        Retrieves the read time of the article.

        Returns:
        - float: The read time of the article.
        """

        return self._access_data('readTime')

    @property
    @DrukarniaElement.type_decorator(int)
    async def number_of_likes(self) -> int:
        """
        Retrieves the number of likes of the article.

        Returns:
        - int: The number of likes of the article.
        """

        return self._access_data('likeNum')

    @property
    @DrukarniaElement.type_decorator(int)
    async def number_of_comments(self) -> int:
        """
        Retrieves the number of comments of the article.

        Returns:
        - int: The number of comments of the article.
        """

        return self._access_data('commentNum')

    @property
    async def article_tags(self) -> Tuple['Tag']:
        """
        Retrieves the tags of the article.

        Returns:
        - Tuple['Tag']: A tuple containing the tags of the article.
        """

        tags = self._access_data('tags', [])

        if any(map(lambda el: isinstance(el, str), tags)):
            warn("Tag's data is incomplete, use collect_data method before in order to obtain whole data.")
            tags = map(lambda el: {'_id': el}, tags)

        return await data2tags(tags, self.session)

    @property
    async def main_article_tag(self) -> 'Tag' or None:
        """
        Retrieves the main tag of the article.

        Returns:
        - 'Tag' or None: The main tag of the article if available, otherwise None.
        """

        main_id = self._access_data('mainTagId', str)
        main_name = self._access_data('mainTag', str)
        main_slug = self._access_data('mainTagSlug', str)

        if not (main_id and main_name and main_slug):
            return None

        main_tag, = await data2tags([{'_id': main_id, 'name': main_name, 'slug': main_slug}], self.session)

        return main_tag

    @property
    @DrukarniaElement.type_decorator(str)
    async def description(self) -> str:
        """
        Retrieves the description of the article.

        Returns:
        - str: The description of the article.
        """

        return self._access_data('description')

    @property
    @DrukarniaElement.type_decorator(str)
    async def seo_title(self) -> str:
        """
        Retrieves the SEO title of the article.

        Returns:
        - str: The SEO title of the article.
        """
        return self._access_data('seoTitle')

    @property
    @DrukarniaElement.type_decorator(str)
    async def title(self) -> str:
        """
        Retrieves the title of the article.

        Returns:
        - str: The title of the article.
        """
        return self._access_data('title')

    @property
    @DrukarniaElement.type_decorator(str)
    async def article_id(self) -> str:
        """
        Retrieves the ID of the article.

        Returns:
        - str: The ID of the article.
        """
        return self._access_data('_id')

    @property
    @DrukarniaElement.type_decorator(str)
    async def slug(self) -> str:
        """
        Retrieves the slug of the article.

        Returns:
        - str: The slug of the article.
        """
        return self._access_data('slug')

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Article':
        """
        Creates an Article instance from records.

        Parameters:
        - session (ClientSession): The client session.
        - new_data (dict): The new data to create the Article instance.

        Returns:
        - 'Article': The created Article instance.
        """

        new_article = Article(session=session)
        new_article._update_data(new_data)

        return new_article

    def __hash__(self) -> int:
        """
        Returns the hash value of the Article object.

        Returns:
        - int: The hash value of the Article object.
        """
        return hash(self.article_id or self.slug)



from datetime import datetime

from aiohttp import ClientSession

from drukarnia_api.objects.base_object import DrukarniaElement
from drukarnia_api.shortcuts import data2articles, data2tags, data2authors

from typing import TYPE_CHECKING, Tuple, Dict

if TYPE_CHECKING:   # always False, used for type hints
    from drukarnia_api.objects.article import Article
    from drukarnia_api.objects.author import Author


class Tag(DrukarniaElement):
    def __init__(self, slug_name: str = None, tag_id: str = None, *args, **kwargs):
        """
        Initialize a Tag object.

        Args:
            slug_name (str, optional): The slug name of the tag. Defaults to None.
            tag_id (str, optional): The ID of the tag. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

        # Update the data with slug_name and tag_id
        self._update_data({'slug': slug_name, '_id': tag_id})

    @DrukarniaElement.requires_attributes(['slug'])
    async def get_articles(self, create_articles: bool = True,
                           offset: int = 0, results_per_page: int = 20, n_collect: int = None,
                           **kwargs) -> Tuple['Article'] or Tuple[Dict]:
        """
        Get articles associated with this tag.

        Args:
            create_articles (bool, optional): Whether to create Article objects from the retrieved data.
                                              Defaults to True.
            offset (int, optional): The offset of the result set. Defaults to 0.
            results_per_page (int, optional): The number of results per page. Defaults to 20.
            n_collect (int, optional): The number of articles to collect. Defaults to None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tuple[Article] or Tuple[Dict]: A tuple of Article objects or a tuple of dictionaries containing article data.
        """

        # Make a request to get the followers of the author
        articles = await self.multi_page_request(f'/api/articles/tags/{await self.slug}',
                                                 offset, results_per_page, n_collect, list_key='articles',
                                                 **kwargs)
        if create_articles:
            articles = await data2articles(articles, self.session)

        return articles

    @DrukarniaElement.requires_attributes(['tag_id'])
    async def related_tags(self, create_tags: bool = True) -> Tuple['Tag'] or Tuple[Dict]:
        """
        Get tags related to this tag.

        Args:
            create_tags (bool, optional): Whether to create Tag objects from the retrieved data. Defaults to True.

        Returns:
            Tuple[Tag] or Tuple[Dict]: A tuple of Tag objects or a tuple of dictionaries containing tag data.
        """

        # Make a request to get the followers of the author
        tags = await self.request('get', f'/api/articles/tags/{await self.tag_id}/related?page=1', output='json')
        if create_tags:
            tags = await data2tags(tags, self.session)

        return tags

    @DrukarniaElement.requires_attributes(['tag_id'])
    async def subscribe_tag(self, unsubscribe: bool = False) -> None:
        """
        Subscribe or unsubscribe to/from this tag.

        Args:
            unsubscribe (bool, optional): Whether to unsubscribe from the tag. Defaults to False.
        """
        if unsubscribe:
            await self.request('delete', f'/api/preferences/tags/{await self.tag_id}')
            return None

        await self.request('put', f'/api/preferences/tags/{await self.tag_id}')

    @DrukarniaElement.requires_attributes(['tag_id'])
    async def block_tag(self, unblock: bool = False) -> None:
        """
        Block or unblock this tag.

        Args:
            unblock (bool, optional): Whether to unblock the tag. Defaults to False.
        """
        if unblock:
            await self.request('put', f'/api/preferences/tags/{await self.tag_id}/block', data={"isBlocked": False})
            return None

        await self.request('put', f'/api/preferences/tags/{await self.tag_id}/block', data={"isBlocked": True})

    @DrukarniaElement.requires_attributes(['tag_id'])
    async def related_authors(self, create_tags: bool = True,) -> Tuple['Author'] or Tuple[Dict]:
        """
        Get authors related to this tag.

        Args:
            create_tags (bool, optional): Whether to create Author objects from the retrieved data. Defaults to True.

        Returns:
            Tuple[Author] or Tuple[Dict]: A tuple of Author objects or a tuple of dictionaries containing author data.
        """

        # Make a request to get the followers of the author
        authors = await self.request('get', f'/api/users/tags/{await self.tag_id}/related', output='json')
        if create_tags:
            authors = await data2authors(authors, self.session)

        return authors

    @DrukarniaElement.requires_attributes(['slug'])
    async def collect_data(self, return_: bool = False):
        """
        Collect data associated with this tag.

        Args:
            return_ (bool, optional): Whether to return the collected data. Defaults to False.

        Returns:
            dict or None: The collected data or None if return_ is False.
        """
        result = await self.request('get', f'/api/articles/tags/{await self.slug}?page=1', output='json')

        if result.get('articles', None):
            del result['articles']

        self._update_data(result)

        if return_:
            return result

    @property
    @DrukarniaElement.type_decorator(str)
    async def slug(self):
        """
        Get the slug property of the Tag.

        Returns:
            str: The slug property of the Tag.
        """
        return self._access_data('slug')

    @property
    @DrukarniaElement.type_decorator(datetime)
    async def created_at(self):
        """
        Get the created_at property of the Tag.

        Returns:
            datetime: The created_at property of the Tag.
        """
        return self._access_data('createdAt')

    @property
    @DrukarniaElement.type_decorator(bool)
    async def default(self):
        """
        Get the default property of the Tag.

        Returns:
            bool: The default property of the Tag.
        """
        return self._access_data('default')

    @property
    @DrukarniaElement.type_decorator(int)
    async def mentions_num(self):
        """
        Get the mentions_num property of the Tag.

        Returns:
            int: The mentions_num property of the Tag.
        """
        return self._access_data('mentionsNum')

    @property
    @DrukarniaElement.type_decorator(str)
    async def name(self):
        """
        Get the name property of the Tag.

        Returns:
            str: The name property of the Tag.
        """
        return self._access_data('name')

    @property
    @DrukarniaElement.type_decorator(str)
    async def tag_id(self):
        """
        Get the _id property of the Tag.

        Returns:
            str: The _id property of the Tag.
        """
        return self._access_data('_id')

    @property
    @DrukarniaElement.type_decorator(dict)
    async def relationships(self):
        """
        Get the relationships property of the Tag.

        Returns:
            dict: The relationships property of the Tag.
        """
        return self._access_data('relationships')

    @staticmethod
    async def from_records(session: ClientSession, new_data: dict) -> 'Tag':
        """
        Create a new Tag instance from records.

        Args:
            session (ClientSession): A session object for making HTTP requests.
            new_data (dict): Data to update the new Tag instance with.

        Returns:
            Tag: A new instance of the Tag class.
        """
        new_tag = Tag(session=session)
        new_tag._update_data(new_data)

        return new_tag



from aiohttp import ClientResponse
from typing import Any, Dict, List
from drukarnia_api.network.exceptions import DrukarniaAPIError

import json
import inspect


async def _from_response(response: ClientResponse, output: str or List[str] or None) -> Any:
    """
    Extracts data from the aiohttp ClientResponse based on the provided output.
    If the response status code is 400 or above, raises a DrukarniaAPIError.

    Parameters:
        response (ClientResponse): The aiohttp ClientResponse object.
        output (str or List[str] or None): A string, a list of strings, or None.
            If a string, it should be the attribute name to extract from the response.
            If a list of strings, it should be a list of attribute names to extract from the response.
            If None, returns an empty list.

    Returns:
        Any: The extracted data based on the provided output.
    """

    # for some reason Drukarnia uses 2xx (> 201) codes for errors
    if int(response.status) > 201:
        data = await response.json()
        raise DrukarniaAPIError(data['message'], response.status,
                                response.request_info.method, str(response.request_info.url))

    if output is None:
        return []

    elif isinstance(output, str):
        data = await _from_response(response, [output])
        return data[0]

    data = []
    for func_name in output:
        attr = getattr(response, func_name)

        if inspect.iscoroutinefunction(attr):
            data.append(await attr())

        elif callable(attr):
            data.append(attr())

        else:
            data.append(attr)

    return data


async def to_json(data: Any):
    """
    Converts the provided data into a JSON string if it's a dictionary or a list.

    Parameters:
        data (Any): The data to be converted to JSON.

    Returns:
        Any: If the data is a dictionary or a list, it returns the JSON string representation.
            Otherwise, it returns the data unchanged.
    """

    if isinstance(data, Dict) or isinstance(data, List):
        return json.dumps(data)

    return data
